@@ -1513,6 +1513,9 @@ static int voicecall_dial(struct ofono_voicecall *vc, const char *number,
 	if (g_slist_length(vc->call_list) >= MAX_VOICE_CALLS)
 		return -EPERM;
 
+	if (valid_ussd_string(number, vc->call_list != NULL))
+		return -EINVAL;
+
 	if (!valid_long_phone_number_format(number))
 		return -EINVAL;
 
@@ -2355,6 +2358,46 @@ error:
 		g_free(v);
 }
 
+void ofono_voicecall_mpty_hint(struct ofono_voicecall *vc, unsigned int ids)
+{
+	GSList *old;
+	GSList *l;
+
+	DBG("ids: %u", ids);
+
+	/* id of 0 is never valid for a call */
+	if (ids & 0x1)
+		return;
+
+	/* Ignore the hint if there's nothing to do */
+	if (__builtin_popcount(ids) < 2 && vc->multiparty_list == NULL)
+		return;
+
+	old = vc->multiparty_list;
+	vc->multiparty_list = NULL;
+
+	for (l = vc->call_list; l; l = l->next) {
+		struct voicecall *v = l->data;
+
+		if (ids & (1 << v->call->id))
+			vc->multiparty_list =
+				g_slist_prepend(vc->multiparty_list, v);
+	}
+
+	if (vc->multiparty_list)
+		vc->multiparty_list = g_slist_reverse(vc->multiparty_list);
+
+	if (g_slist_length(vc->multiparty_list) == 1) {
+		ofono_error("Created multiparty list length is 1"
+				", which would indicate a bug in the driver"
+				" or the remote device");
+		vc->multiparty_list = NULL;
+	}
+
+	voicecalls_multiparty_changed(old, vc->multiparty_list);
+	g_slist_free(old);
+}
+
 static void send_ciev_after_swap_callback(const struct ofono_error *error,
 								void *data)
 {
@@ -2834,6 +2877,7 @@ static void sim_state_watch(enum ofono_sim_state new_state, void *user)
 						read_sim_ecc_numbers, vc, NULL);
 		break;
 	case OFONO_SIM_STATE_NOT_PRESENT:
+	case OFONO_SIM_STATE_RESETTING:
 		/* TODO: Must release all non-emergency calls */
 
 		if (vc->sim_context) {
@@ -3263,6 +3307,10 @@ static void emulator_chld_cb(struct ofono_emulator *em,
 
 		ofono_emulator_send_info(em, buf, TRUE);
 		result.type = OFONO_ERROR_TYPE_NO_ERROR;
+
+		__ofono_emulator_slc_condition(em,
+					OFONO_EMULATOR_SLC_CONDITION_CHLD);
+
 		break;
 
 	case OFONO_EMULATOR_REQUEST_TYPE_QUERY:
@@ -3868,15 +3916,21 @@ void __ofono_voicecall_set_alpha_and_icon_id(struct ofono_voicecall *vc,
 	vc->dial_req = req;
 
 	vc->flags |= VOICECALL_FLAG_STK_MODEM_CALLSETUP;
+
+	DBG("%p, %p", vc, vc->dial_req);
 }
 
 void __ofono_voicecall_clear_alpha_and_icon_id(struct ofono_voicecall *vc)
 {
-	g_free(vc->dial_req->message);
-	vc->dial_req->message = NULL;
+	DBG("%p, %p", vc, vc->dial_req);
 
-	g_free(vc->dial_req);
-	vc->dial_req = NULL;
+	if (vc->dial_req) {
+		g_free(vc->dial_req->message);
+		vc->dial_req->message = NULL;
+
+		g_free(vc->dial_req);
+		vc->dial_req = NULL;
+	}
 
 	vc->flags &= ~VOICECALL_FLAG_STK_MODEM_CALLSETUP;
 }
