@@ -38,9 +38,18 @@ static const char *simple_deliver = "07911326040000F0"
 		"040B911346610089F60000208062917314480CC8F71D14969741F977FD07";
 static const char *alnum_sender = "0791447758100650"
 		"040DD0F334FC1CA6970100008080312170224008D4F29CDE0EA7D9";
+static const char *unicode_deliver = "04819999990414D0FBFD7EBFDFEFF77BFE1E001"
+		"9512090801361807E00DC00FC00C400E400D600F600C500E500D800F800C"
+		"600E600C700E700C900E900CA00EA00DF003100320033003400350036003"
+		"7003800390030002000540068006900730020006D0065007300730061006"
+		"7006500200069007300200036003300200075006E00690063006F0064006"
+		"5002000630068006100720073002E";
 static const char *simple_submit = "0011000B916407281553F80000AA"
 		"0AE8329BFD4697D9EC37";
 
+static const char *simple_mwi = "07913366002020F8040B913366600600F100C8318070"
+				"6174148000";
+
 static void print_scts(struct sms_scts *scts, const char *prefix)
 {
 	time_t ts;
@@ -359,6 +368,38 @@ static void test_deliver_encode(void)
 	g_assert(strcmp(alnum_sender, encoded_pdu) == 0);
 
 	g_free(encoded_pdu);
+
+	/* test unicode_deliver*/
+	decoded_pdu = decode_hex(unicode_deliver, -1, &pdu_len, 0);
+	g_assert(decoded_pdu);
+	g_assert(pdu_len == (long)strlen(unicode_deliver) / 2);
+
+	ret = sms_decode(decoded_pdu, pdu_len, FALSE, 149, &sms);
+
+	g_free(decoded_pdu);
+
+	g_assert(ret);
+	g_assert(sms.type == SMS_TYPE_DELIVER);
+
+	ret = sms_encode(&sms, &encoded_pdu_len, &encoded_tpdu_len, pdu);
+
+	if (g_test_verbose()) {
+		int i;
+
+		for (i = 0; i < encoded_pdu_len; i++)
+			g_print("%02X", pdu[i]);
+		g_print("\n");
+	}
+
+	g_assert(ret);
+	g_assert(encoded_tpdu_len == 149);
+	g_assert(encoded_pdu_len == pdu_len);
+
+	encoded_pdu = encode_hex(pdu, encoded_pdu_len, 0);
+
+	g_assert(strcmp(unicode_deliver, encoded_pdu) == 0);
+
+	g_free(encoded_pdu);
 }
 
 static void test_simple_submit(void)
@@ -471,6 +512,107 @@ static void test_submit_encode(void)
 	g_free(encoded_pdu);
 }
 
+static void test_simple_mwi(void)
+{
+	struct sms sms;
+	unsigned char *decoded_pdu;
+	long pdu_len;
+	gboolean ret;
+	enum sms_mwi_type type;
+	gboolean active;
+	gboolean discard;
+
+	decoded_pdu = decode_hex(simple_mwi, -1, &pdu_len, 0);
+
+	g_assert(decoded_pdu);
+	g_assert(pdu_len == (long)strlen(simple_mwi) / 2);
+
+	ret = sms_decode(decoded_pdu, pdu_len, FALSE, 19, &sms);
+
+	g_free(decoded_pdu);
+
+	g_assert(ret);
+	g_assert(sms.type == SMS_TYPE_DELIVER);
+
+	if (g_test_verbose())
+		dump_details(&sms);
+
+	g_assert(sms.sc_addr.number_type == SMS_NUMBER_TYPE_INTERNATIONAL);
+	g_assert(sms.sc_addr.numbering_plan == SMS_NUMBERING_PLAN_ISDN);
+	g_assert(strcmp(sms.sc_addr.address, "33660002028") == 0);
+
+	g_assert(sms.deliver.oaddr.number_type ==
+			SMS_NUMBER_TYPE_INTERNATIONAL);
+	g_assert(sms.deliver.oaddr.numbering_plan ==
+			SMS_NUMBERING_PLAN_ISDN);
+	g_assert(strcmp(sms.deliver.oaddr.address, "33660660001") == 0);
+
+	g_assert(sms.deliver.pid == 0);
+	g_assert(sms.deliver.dcs == 200);
+
+	g_assert(sms.deliver.scts.year == 13);
+	g_assert(sms.deliver.scts.month == 8);
+	g_assert(sms.deliver.scts.day == 7);
+	g_assert(sms.deliver.scts.hour == 16);
+	g_assert(sms.deliver.scts.minute == 47);
+	g_assert(sms.deliver.scts.second == 41);
+	g_assert(sms.deliver.scts.timezone == 8);
+
+	g_assert(sms.deliver.udl == 0);
+
+	if (sms.deliver.udhi) {
+		struct sms_udh_iter iter;
+		enum sms_iei iei;
+
+		ret = sms_udh_iter_init(&sms, &iter);
+		g_assert(ret);
+
+		while ((iei = sms_udh_iter_get_ie_type(&iter)) !=
+				SMS_IEI_INVALID) {
+			switch (iei) {
+			case SMS_IEI_ENHANCED_VOICE_MAIL_INFORMATION:
+			{
+				unsigned char evm_iei[140];
+				sms_udh_iter_get_ie_data(&iter, evm_iei);
+				sms_udh_iter_get_ie_length(&iter);
+
+				if (g_test_verbose())
+					g_print("Enhanced Voicemail IEI\n");
+				break;
+			}
+
+			case SMS_IEI_SPECIAL_MESSAGE_INDICATION:
+			{
+				unsigned char special_iei[4];
+
+				sms_udh_iter_get_ie_data(&iter, special_iei);
+				sms_udh_iter_get_ie_length(&iter);
+
+				if (g_test_verbose())
+					g_print("Special Voicemail IEI\n");
+
+				break;
+			}
+
+			default:
+				break;
+			}
+
+			sms_udh_iter_next(&iter);
+		}
+	}
+
+	ret = sms_mwi_dcs_decode(sms.deliver.dcs, &type, NULL, &active,
+					&discard);
+	g_assert(ret);
+
+	if (g_test_verbose()) {
+		g_print("Type: %d, Active: %d, Discard: %d\n",
+				type, active, discard);
+
+	}
+}
+
 struct sms_charset_data {
 	char *pdu;
 	int data_len;
@@ -850,13 +992,13 @@ static void test_ems_udh(gconstpointer data)
 	for (i = 0; test->formats[i].len; i++) {
 		if (g_test_verbose()) {
 			int j;
-			unsigned char data[4];
+			unsigned char ie_data[4];
 
-			sms_udh_iter_get_ie_data(&iter, data);
+			sms_udh_iter_get_ie_data(&iter, ie_data);
 
 			g_print("Header:\n");
 			for (j = 0; j < sms_udh_iter_get_ie_length(&iter); j++)
-				g_print("0x%02x ", data[j]);
+				g_print("0x%02x ", ie_data[j]);
 
 			g_print("\n");
 		}
@@ -1236,6 +1378,10 @@ static const char *cbs2 = "0110003201114679785E96371A8D46A3D168341A8D46A3D1683"
 	"41A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D168"
 	"341A8D46A3D168341A8D46A3D168341A8D46A3D168341A8D46A3D100";
 
+static const char *cbs3 = "001000000111E280604028180E888462C168381E90886442A95"
+	"82E988C66C3E9783EA09068442A994EA8946AC56AB95EB0986C46ABD96EB89C6EC7EBF"
+	"97EC0A070482C1A8FC8A472C96C3A9FD0A8744AAD5AAFD8AC76CB05";
+
 static void test_cbs_encode_decode(void)
 {
 	unsigned char *decoded_pdu;
@@ -1389,6 +1535,58 @@ static void test_cbs_assembly(void)
 	cbs_assembly_free(assembly);
 }
 
+static void test_cbs_padding_character(void)
+{
+	unsigned char *decoded_pdu;
+	long pdu_len;
+	gboolean ret;
+	struct cbs cbs;
+	GSList *l;
+	char iso639_lang[3];
+	char *utf8;
+
+	decoded_pdu = decode_hex(cbs3, -1, &pdu_len, 0);
+
+	g_assert(decoded_pdu);
+	g_assert(pdu_len == 88);
+
+	ret = cbs_decode(decoded_pdu, pdu_len, &cbs);
+
+	g_free(decoded_pdu);
+
+	g_assert(ret);
+
+	g_assert(cbs.gs == CBS_GEO_SCOPE_CELL_IMMEDIATE);
+	g_assert(cbs.message_code == 1);
+	g_assert(cbs.update_number == 0);
+	g_assert(cbs.message_identifier == 0);
+	g_assert(cbs.dcs == 1);
+	g_assert(cbs.max_pages == 1);
+	g_assert(cbs.page == 1);
+
+	l = g_slist_append(NULL, &cbs);
+
+	utf8 = cbs_decode_text(l, iso639_lang);
+
+	g_assert(utf8);
+
+	if (g_test_verbose()) {
+		g_printf("%s\n", utf8);
+		if (iso639_lang[0] == '\0')
+			g_printf("Lang: Unspecified\n");
+		else
+			g_printf("Lang: %s\n", iso639_lang);
+	}
+
+	g_assert(strcmp(utf8, "b£$¥èéùìòÇ\x0AØø\x0DÅåΔ_ΦΓΛΩΠΨΣΘΞÆæßÉ !\"#¤"
+				"\x25&'()*+,-./0123456789:;<=>?¡ABCDEFGHIJKLM"
+				"NOPQRSTUVWXYZÄÖ") == 0);
+	g_assert(strcmp(iso639_lang, "en") == 0);
+
+	g_free(utf8);
+	g_slist_free(l);
+}
+
 static const char *ranges[] = { "1-5, 2, 3, 600, 569-900, 999",
 				"0-20, 33, 44, 50-60, 20-50, 1-5, 5, 3, 5",
 				NULL };
@@ -1621,6 +1819,7 @@ int main(int argc, char **argv)
 	g_test_add_func("/testsms/Test Deliver Encode", test_deliver_encode);
 	g_test_add_func("/testsms/Test Simple Submit", test_simple_submit);
 	g_test_add_func("/testsms/Test Submit Encode", test_submit_encode);
+	g_test_add_func("/testsms/Test Simple MWI", test_simple_mwi);
 
 	g_test_add_data_func("/testsms/Test "
 		"GSM 7 bit Default Alphabet Decode",
@@ -1679,6 +1878,9 @@ int main(int argc, char **argv)
 			test_cbs_encode_decode);
 	g_test_add_func("/testsms/Test CBS Assembly", test_cbs_assembly);
 
+	g_test_add_func("/testsms/Test CBS Padding Character",
+			test_cbs_padding_character);
+
 	g_test_add_func("/testsms/Range minimizer", test_range_minimizer);
 
 	g_test_add_func("/testsms/Status Report Assembly", test_sr_assembly);
